package ratelimit

import (
	"bufio"
	"container/heap"
	"fmt"
	"hash/fnv"
	"io"
	"net"
	"sync"
	"time"

	"github.com/bmkessler/streamstats"
)

const linearCountingP = 3

// Entry with separate success/failure counters and a cardinality
// estimator.
type cardinalityEntry struct {
	entry

	countOk     int64
	countFail   int64
	cardinality *streamstats.LinearCounting
}

func (e *cardinalityEntry) Reset(expiry int64) {
	e.countOk = 0
	e.countFail = 0
	e.expiry = expiry
	e.cardinality = streamstats.NewLinearCounting(linearCountingP, fnv.New64a())
}

// CardinalityParams establish the criteria used to evaluate the
// behavior of a specific key. By manipulating thresholds and weights
// it should be possible to discriminate between distributed
// brute-forcers, IP proxies with real users, etc.
type CardinalityParams struct {
	OkWeight   float64 `yaml:"ok_weight"`
	FailWeight float64 `yaml:"fail_weight"`

	MaxRequestRate    float64 `yaml:"max_request_rate"`
	MinRequestOkRatio float64 `yaml:"min_request_ok_ratio"`
	MaxCardinality    int     `yaml:"max_cardinality"`
}

func DefaultCardinalityParams() CardinalityParams {
	return CardinalityParams{
		OkWeight:          1.0,
		FailWeight:        10.0,
		MaxRequestRate:    0.2,
		MinRequestOkRatio: 0.9,
		MaxCardinality:    3,
	}
}

// Score evaluates the current values against the given
// CardinalityParams, and returns a cumulative category score.
func (e *cardinalityEntry) Score(now int64, params CardinalityParams) int {
	elapsed := float64(e.expiry - now)
	// Bound elapsed to 1 second to avoid NaNs.
	if elapsed < 1 {
		elapsed = 1
	}

	reqRate := float64(e.countFail+e.countOk) / elapsed
	isHighRequestRate := reqRate > params.MaxRequestRate

	ratioOk := ((1.0 + params.OkWeight*float64(e.countOk)) /
		(1.0 + params.FailWeight*float64(e.countFail) + params.OkWeight*float64(e.countOk)))
	isHighFailureRate := ratioOk < params.MinRequestOkRatio

	distinct := e.cardinality.Distinct()
	isHighCardinality := distinct > uint64(params.MaxCardinality)

	// One point for each anomaly criteria.
	return b2i(isHighRequestRate) + b2i(isHighFailureRate) + b2i(isHighCardinality)
}

// Cardinality counts events by key, and keeps track of the
// approximate cardinality of the event values, as well as the
// associated event status (success or failure).
//
// It works pretty much the same as a Counter, but stores
// CriticalityEntry objects instead.
type cardinality struct {
	pq      priorityQueue
	byKey   map[key]*cardinalityEntry
	maxSize int

	interval int64

	gc *gc
}

func newCardinality(interval, maxSize int, gcParams GCParams) *cardinality {
	c := &cardinality{
		byKey:    make(map[key]*cardinalityEntry),
		maxSize:  maxSize,
		interval: int64(interval),
		gc:       newGC(gcParamsForMaxSize(maxSize, gcParams)),
	}
	heap.Init(&c.pq)
	return c
}

func (c *cardinality) Get(now int64, key key) *cardinalityEntry {
	if entry, ok := c.byKey[key]; ok && entry.expiry > now {
		return entry
	}
	return nil
}

func (c *cardinality) Inc(now int64, key key, value string, status bool) *cardinalityEntry {
	c.gc.run(c, now)

	entry, ok := c.byKey[key]
	if !ok {
		if c.maxSize > 0 && len(c.byKey) >= c.maxSize {
			return nil
		}

		entry = new(cardinalityEntry)
		entry.Reset(now + c.interval)
		entry.key = key

		c.byKey[key] = entry
		heap.Push(&c.pq, entry)
	} else if entry.expiry < now {
		entry.Reset(now + c.interval)
		c.pq.update(entry)
	}

	entry.cardinality.Add([]byte(value))

	if status {
		entry.countOk++
	} else {
		entry.countFail++
	}

	return entry
}

func (c *cardinality) Len() int { return len(c.byKey) }

func (c *cardinality) Interval() int64 { return c.interval }

func (c *cardinality) Pop() any { return heap.Pop(&c.pq) }

func (c *cardinality) Push(entry any) { heap.Push(&c.pq, entry) }

func (c *cardinality) DeleteByKey(key key) { delete(c.byKey, key) }

func (c *cardinality) Dump(w io.Writer) {
	bw := bufio.NewWriter(w)
	now := time.Now().Unix()
	for _, entry := range c.byKey {
		if now < entry.expiry {
			fmt.Fprintf(bw, "%s %d %d %d\n", net.IP(entry.key[:]).String(), entry.countOk, entry.countFail, entry.cardinality.Distinct())
		}
	}
	bw.Flush()
}

// A cardinalityBlocker can block peers for a given amount of time,
// based on the result of the score evaluation of a Cardinality
// instance. Instead of using an accurate rate counter like Blocker,
// it uses CardinalityEntry's qps approximation.
type cardinalityBlockerBase struct {
	block          *counter
	cardinality    *cardinality
	params         CardinalityParams
	scoreThreshold int
}

func newCardinalityBlockerBase(interval, blockInterval, maxSize, scoreThreshold int, params CardinalityParams, gcParams GCParams) *cardinalityBlockerBase {
	return &cardinalityBlockerBase{
		block:          newCounter(blockInterval, maxSize, gcParams),
		cardinality:    newCardinality(interval, maxSize, gcParams),
		params:         params,
		scoreThreshold: scoreThreshold,
	}
}

func (b *cardinalityBlockerBase) Allow(now int64, key key) bool {
	value := b.block.Get(now, key)
	return value == 0
}

func (b *cardinalityBlockerBase) Inc(now int64, key key, value string, status bool) bool {
	entry := b.cardinality.Inc(now, key, value, status)

	if entry.Score(now, b.params) >= b.scoreThreshold {
		b.block.Inc(now, key)
		return true
	}

	return false
}

// CardinalityBlocker is a cardinality-based blocker that is safe to
// call from multiple goroutines, and accepts keys as net.IP addresses
// in its API.
//
// Blocking happens based on multiple criteria including traffic rate,
// value stream cardinality, and success/failure ratio. Check out the
// CardinalityParams documentation for more details.
type CardinalityBlocker struct {
	*cardinalityBlockerBase

	mx     sync.Mutex
	skip   cidrList
	ipMask *ipMask
}

func NewCardinalityBlocker(interval, blockInterval, scoreThreshold int, options ...Option) *CardinalityBlocker {
	opts := applyOptions(options)

	return &CardinalityBlocker{
		cardinalityBlockerBase: newCardinalityBlockerBase(
			interval, blockInterval, opts.maxSize, scoreThreshold, opts.cparams, opts.gcParams),
		skip:   opts.skip,
		ipMask: opts.ipMask,
	}
}

func (b *CardinalityBlocker) toKey(ip net.IP) key {
	return ipToKey(b.ipMask.Mask(ip))
}

// Allow returns true if the incoming request is allowed.
func (b *CardinalityBlocker) Allow(now int64, ip net.IP) bool {
	if b.skip.Contains(ip) {
		return true
	}

	b.mx.Lock()
	defer b.mx.Unlock()

	return b.cardinalityBlockerBase.Allow(now, b.toKey(ip))
}

// Inc increments the rate limit counter for the specified peer, and
// records 'value' via the cardinality estimator. If this causes the
// peer to be blocked, the function will return true.
func (b *CardinalityBlocker) Inc(now int64, ip net.IP, value string, status bool) bool {
	if b.skip.Contains(ip) {
		return false
	}

	b.mx.Lock()
	defer b.mx.Unlock()

	return b.cardinalityBlockerBase.Inc(now, b.toKey(ip), value, status)
}

func b2i(b bool) int {
	if b {
		return 1
	}
	return 0
}
