package ratelimit

import (
	"fmt"
	"testing"
	"time"
)

func TestCardinality(t *testing.T) {
	c := newCardinality(10, 0, DefaultGCParams())

	key1 := testkey()

	var ent *cardinalityEntry
	for i := 0; i < 1000; i++ {
		ent = c.Inc(0, key1, fmt.Sprintf("user%05d", i), true)
	}

	distinct := ent.cardinality.Distinct()
	if distinct != 10 {
		t.Errorf("Distinct() = %d, expected 10", distinct)
	}
}

func TestCardinality_Cleanup(t *testing.T) {
	key1 := testkey()
	key2 := testkey()

	c := newCardinality(10, 0, DefaultGCParams())
	c.Inc(0, key1, "", true)
	c.Inc(1, key1, "", true)
	cleanup(c, 2, time.Now().Add(1*time.Second))

	if c.Get(3, key1) == nil {
		t.Errorf("key1@3: entry is nil, expected non-nil")
	}

	c.Inc(10, key1, "", true)
	c.Inc(99, key2, "", true)
	cleanup(c, 100, time.Now().Add(1*time.Second))

	if e := c.Get(101, key1); e != nil {
		t.Errorf("key1@101: entry=%v, expected nil", e)
	}
	if _, ok := c.byKey[key1]; ok {
		t.Errorf("key1@101: entry is nil but the entry was not cleaned up")
	}

	if c.Get(101, key2) == nil {
		t.Errorf("key2@101: entry is nil, expected non-nil")
	}
}

type tsvEntry struct {
	t     int64
	key   key
	value string
}

func timeseriesValues(key key, duration, rate float64, numValues int) []tsvEntry {
	n := int(rate * duration)
	out := make([]tsvEntry, n)
	for i := 0; i < n; i++ {
		t := int64(float64(i) * duration / float64(n))
		value := fmt.Sprintf("value%06d", i%numValues)
		out[i] = tsvEntry{t: t, key: key, value: value}
	}
	return out
}

func TestCardinalityBlocker(t *testing.T) {
	b := newCardinalityBlockerBase(
		3, 100, 0, 2, CardinalityParams{
			OkWeight:          1.0,
			FailWeight:        10.0,
			MaxRequestRate:    0.1,
			MinRequestOkRatio: 0.9,
			MaxCardinality:    3,
		}, DefaultGCParams(),
	)

	// Use high qps to populate the timeseries with enough values,
	// and to ensure that the highRequestRate check always
	// triggers.

	// High cardinality (blocked).
	key1 := testkey()
	for _, entry := range timeseriesValues(key1, 3.0, 2000, 1000) {
		b.Inc(entry.t, entry.key, entry.value, true)
	}

	// Low cardinality (not blocked).
	key2 := testkey()
	for _, entry := range timeseriesValues(key2, 3.0, 2000, 1) {
		b.Inc(entry.t, entry.key, entry.value, true)
	}

	// High failure rate, low cardinality (blocked).
	key3 := testkey()
	for i, entry := range timeseriesValues(key3, 3.0, 2000, 1) {
		status := ((i % 10) != 0)
		b.Inc(entry.t, entry.key, entry.value, status)
	}

	if ok := b.Allow(3, key1); ok {
		t.Errorf("key1 (high-c) was not blocked")
	}

	if ok := b.Allow(6, key1); ok {
		t.Errorf("key1 (high-c) block expired too soon")
	}

	if ok := b.Allow(3, key2); !ok {
		t.Errorf("key2 (low-c) was blocked")
	}

	if ok := b.Allow(3, key3); ok {
		t.Errorf("key3 (high-failure) was not blocked")
	}
}

func BenchmarkCardinality_Dense(b *testing.B) {
	b.StopTimer()

	c := newCardinality(10, 0, DefaultGCParams())

	keys := testkeys(100)

	b.StartTimer()

	for i := 0; i < b.N; i++ {
		t := int64(100 * i / b.N)
		c.Inc(t, keys[i%len(keys)], fmt.Sprintf("user%05d", i%1000), true)
	}
}

func BenchmarkCardinality_Sparse(b *testing.B) {
	b.StopTimer()

	c := newCardinality(10, 0, DefaultGCParams())

	b.StartTimer()

	for i := 0; i < b.N; i++ {
		t := int64(100 * i / b.N)
		c.Inc(t, testkey(), "value", true)
	}
}
