package ratelimit

import (
	"bufio"
	"container/heap"
	"fmt"
	"io"
	"net"
	"time"
)

// The primary key type is just a fixed-length net.IP representation.
// This makes it a pure value type, which helps a lot with Go GC given
// the usage of this type as a map[] key.
type key [16]byte

// Base entry that implements the queueItem interface for
// priorityQueue (and also stores a key).
type entry struct {
	key    key
	index  int
	expiry int64
}

func (e entry) Key() key        { return e.key }
func (e entry) Index() int      { return e.index }
func (e entry) Expiry() int64   { return e.expiry }
func (e *entry) SetIndex(i int) { e.index = i }

// Entry with a simple counter.
type counterEntry struct {
	entry

	count int64
}

func (e *counterEntry) Reset(expiry int64) {
	e.count = 0
	e.expiry = expiry
}

// Counter counts events by key, accumulating values that occur within
// a set time interval.
//
// The structure is not maximally space-efficient since keys are
// stored twice (in the map, and in the Entry itself), all Entry
// pointers are duplicated, and there's overhead from maintaining the
// 'index' element as part of the priorityQueue.
type counter struct {
	pq      priorityQueue
	byKey   map[key]*counterEntry
	maxSize int

	interval int64

	gc *gc
}

func newCounter(interval, maxSize int, gcParams GCParams) *counter {
	c := &counter{
		byKey:    make(map[key]*counterEntry),
		maxSize:  maxSize,
		interval: int64(interval),
		gc:       newGC(gcParamsForMaxSize(maxSize, gcParams)),
	}
	heap.Init(&c.pq)
	return c
}

func (c *counter) Get(now int64, key key) int64 {
	if entry, ok := c.byKey[key]; ok && entry.expiry > now {
		return entry.count
	}
	return 0
}

func (c *counter) Inc(now int64, key key) int64 {
	c.gc.run(c, now)

	entry, ok := c.byKey[key]
	if !ok {
		if c.maxSize > 0 && len(c.byKey) >= c.maxSize {
			return -1
		}

		entry = new(counterEntry)
		entry.Reset(now + c.interval)
		entry.key = key

		c.byKey[key] = entry
		heap.Push(&c.pq, entry)
	} else if entry.expiry < now {
		entry.Reset(now + c.interval)
		c.pq.update(entry)
	}

	entry.count++

	return entry.count
}

func (c *counter) Len() int { return len(c.byKey) }

func (c *counter) Interval() int64 { return c.interval }

func (c *counter) Pop() any { return heap.Pop(&c.pq) }

func (c *counter) Push(entry any) { heap.Push(&c.pq, entry) }

func (c *counter) DeleteByKey(key key) { delete(c.byKey, key) }

func (c *counter) Dump(w io.Writer) {
	bw := bufio.NewWriter(w)
	now := time.Now().Unix()
	for _, entry := range c.byKey {
		if now < entry.expiry {
			fmt.Fprintf(bw, "%s %d\n", net.IP(entry.key[:]).String(), entry.count)
		}
	}
	bw.Flush()
}
