package ratelimit

import "time"

// Garbage collection target.
type gcTarget interface {
	// Number of items in the collection.
	Len() int

	// Time horizon of the collection.
	Interval() int64

	// Priority queue manipulation.
	Pop() any
	Push(any)
	DeleteByKey(key)
}

// GCParams specifies garbage collection tuning parameters.
type GCParams struct {
	// Minimum size of the target collection required to trigger
	// GC at all. Set to 0 to automatically tune it at a fixed
	// percentage of MaxSize, if specified.
	MinSize int

	// Maximum GC execution time.
	Timeout time.Duration

	// Minimum delay between GC.
	MinDelay time.Duration

	// Maximum delay between GC.
	MaxDelay time.Duration

	// Target GC cleanup ratio. Anything above this collapses to
	// MinDelay.
	TargetRatio float64
}

func DefaultGCParams() GCParams {
	// The default GC parameters provide a maximum GC latency of
	// 50ms, and will scale up to a 20% busy cycle.
	return GCParams{
		MinSize:     0,
		Timeout:     50 * time.Millisecond,
		MinDelay:    200 * time.Millisecond,
		MaxDelay:    300 * time.Second,
		TargetRatio: 0.01,
	}
}

func gcParamsForMaxSize(maxSize int, params GCParams) GCParams {
	if params.MinSize == 0 && maxSize > 0 {
		params.MinSize = maxSize / 10
	}
	return params
}

type gc struct {
	params GCParams

	// State
	delay  time.Duration
	lastGC time.Time
}

func newGC(params GCParams) *gc {
	return &gc{params: params}
}

// Run garbage collection on a target.
func (gc *gc) run(target gcTarget, logicalNow int64) {
	n := target.Len()
	if n < gc.params.MinSize {
		return
	}

	now := time.Now()

	if gc.lastGC.Add(gc.delay).Before(now) {
		return
	}

	// Run GC for a small time interval.
	deadline := now.Add(gc.params.Timeout)
	count, partial := cleanup(target, logicalNow, deadline)

	if partial {
		gc.delay = gc.params.MinDelay
	} else {
		d := float64(gc.params.MaxDelay) - float64(gc.params.MinDelay)
		p := float64(count) / float64(n) / gc.params.TargetRatio
		if p > 1.0 {
			p = 1.0
		}
		gc.delay = gc.params.MinDelay + time.Duration(d*(1.0-p))
	}

	gc.lastGC = now
}

type gcEntry interface {
	Key() key
	Expiry() int64
}

// Cleanup some items (time-limited). Note the two timestamps: the
// first is logical time, related to map expiry times, the second is a
// real time deadline to limit execution.
//
// Returns the number of deleted items, and whether execution was
// interrupted due to reaching the deadline.
func cleanup(target gcTarget, now int64, deadline time.Time) (i int, partial bool) {
	cutoff := now - (2 * target.Interval())

	for target.Len() > 0 {
		// Only check the deadline every few iterations.
		if i%100 == 99 && time.Now().After(deadline) {
			partial = true
			break
		}

		entry := target.Pop().(gcEntry)
		if entry == nil {
			break
		}

		if entry.Expiry() > cutoff {
			target.Push(entry)
			break
		}

		target.DeleteByKey(entry.Key())
		i++
	}

	return
}
