package ratelimit

import (
	"bytes"
	"log"
	"net"
)

func ipToKey(ip net.IP) key {
	return key(ip.To16()[:16])
}

func parseCIDR(s string) (*net.IPNet, error) {
	_, ipnet, err := net.ParseCIDR(s)
	if err == nil {
		return ipnet, nil
	}

	ip := net.ParseIP(s)
	if ip == nil {
		return nil, err
	}
	return &net.IPNet{
		IP:   ip,
		Mask: net.CIDRMask(128, 128),
	}, nil
}

type cidrList []*net.IPNet

func parseCIDRList(l []string) cidrList {
	out := make([]*net.IPNet, 0, len(l))
	for _, s := range l {
		ipnet, err := parseCIDR(s)
		if err != nil {
			log.Printf("Warning: '%s': %v", s, err)
			continue
		}
		out = append(out, ipnet)
	}
	return out
}

func (l cidrList) Contains(ip net.IP) bool {
	for _, ipnet := range l {
		if ipnet.Contains(ip) {
			return true
		}
	}
	return false
}

var v4prefix = []byte{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255}

func isIPv4(ip net.IP) bool {
	return len(ip) == 4 || bytes.HasPrefix(ip, v4prefix)
}

type ipMask struct {
	v4 net.IPMask
	v6 net.IPMask
}

func newIPMask(v4bits, v6bits int) *ipMask {
	return &ipMask{
		v4: net.CIDRMask(v4bits, 32),
		v6: net.CIDRMask(v6bits, 128),
	}
}

func (m *ipMask) Mask(ip net.IP) net.IP {
	if isIPv4(ip) {
		return ip.Mask(m.v4)
	}
	return ip.Mask(m.v6)
}
