package ratelimit

type options struct {
	maxSize  int
	skip     cidrList
	cparams  CardinalityParams
	gcParams GCParams
	ipMask   *ipMask
}

func defaultOptions() options {
	return options{
		skip:     parseCIDRList([]string{"127.0.0.1", "::1"}),
		cparams:  DefaultCardinalityParams(),
		gcParams: DefaultGCParams(),
		ipMask:   newIPMask(32, 64),
	}
}

func applyOptions(opts []Option) options {
	o := defaultOptions()
	for _, opt := range opts {
		opt(&o)
	}
	return o
}

type Option func(*options)

// WithMaxSize sets the (optional) maximum size of a collection. When
// a collection reaches this size, it will stop tracking additional
// entries.
func WithMaxSize(maxSize int) Option {
	return func(o *options) {
		o.maxSize = maxSize
	}
}

// WithSkipRanges sets the list of IP ranges (specified as either bare
// IP addresses, or CIDR ranges) to ignore when applying access
// controls.
func WithSkipRanges(skip []string) Option {
	return func(o *options) {
		o.skip = append(o.skip, parseCIDRList(skip)...)
	}
}

// WithCardinalityParams sets custom cardinality evaluation parameters
// for CardinalityBlocker.
func WithCardinalityParams(params CardinalityParams) Option {
	return func(o *options) {
		o.cparams = params
	}
}

// WithGCParams allow tuning garbage collection parameters of containers.
func WithGCParams(gcParams GCParams) Option {
	return func(o *options) {
		o.gcParams = gcParams
	}
}

// WithIPMask sets the mask used for IPv4/IPv6 addresses, expressed as
// the number of bits set to 1 (i.e. CIDR range notation).
func WithIPMask(v4bits, v6bits int) Option {
	return func(o *options) {
		o.ipMask = newIPMask(v4bits, v6bits)
	}
}
