package ratelimit

import (
	"net"
	"sync"
)

type ratelimiterBase struct {
	rate *counter

	threshold int64
}

func newRatelimiterBase(threshold, interval, maxSize int, gcParams GCParams) *ratelimiterBase {
	return &ratelimiterBase{
		rate:      newCounter(interval, maxSize, gcParams),
		threshold: int64(threshold),
	}
}

func (r *ratelimiterBase) Allow(now int64, key key) bool {
	value := r.rate.Inc(now, key)
	return value < r.threshold
}

// RateLimiter is the basic rate limiter. It is safe to call from
// multiple goroutines.
type RateLimiter struct {
	*ratelimiterBase

	mx     sync.Mutex
	skip   cidrList
	ipMask *ipMask
}

func NewRateLimiter(threshold, interval int, options ...Option) *RateLimiter {
	opts := applyOptions(options)

	return &RateLimiter{
		ratelimiterBase: newRatelimiterBase(
			threshold, interval, opts.maxSize, opts.gcParams),
		skip:   opts.skip,
		ipMask: opts.ipMask,
	}
}

func (r *RateLimiter) Allow(now int64, ip net.IP) bool {
	if r.skip.Contains(ip) {
		return true
	}

	r.mx.Lock()
	defer r.mx.Unlock()

	key := ipToKey(r.ipMask.Mask(ip))
	return r.ratelimiterBase.Allow(now, key)
}

type blockerBase struct {
	rate  *counter
	block *counter

	threshold int64
}

func newBlockerBase(threshold, interval, blockInterval, maxSize int, gcParams GCParams) *blockerBase {
	return &blockerBase{
		rate:      newCounter(interval, maxSize, gcParams),
		block:     newCounter(blockInterval, maxSize, gcParams),
		threshold: int64(threshold),
	}
}

func (b *blockerBase) Allow(now int64, key key) bool {
	value := b.block.Get(now, key)
	return value == 0
}

// Returns if the key was newly blocked or not.
func (b *blockerBase) Inc(now int64, key key) bool {
	value := b.rate.Inc(now, key)
	if value == b.threshold {
		b.block.Inc(now, key)
		return true
	}
	return false
}

// Blocker is the basic blocker type. Safe to call from multiple
// goroutines. It will block for 'blockInterval' duration any peer
// that triggers the rate limit (more than 'threshold' requests over
// 'interval').
//
// There are two ways to operate the blocker, depending on whether you
// choose to call Inc() after Allow() returns false, or not: if you
// call Inc() only after an incoming request was successfully allowed,
// upon expiration of the block a peer will be able to make new
// requests again. If, instead, you always call Inc(), the rate
// limiter will continue to increment even while the block is active:
// as a consequence, the block might be renewed right away.
//
// In the latter scenario, you can't reliably use the Inc() return
// value to be notified when a new block is added (e.g. for logging),
// as it will trigger regardless of whether there is an active block.
type Blocker struct {
	*blockerBase

	mx     sync.Mutex
	skip   cidrList
	ipMask *ipMask
}

func NewBlocker(threshold, interval, blockInterval int, options ...Option) *Blocker {
	opts := applyOptions(options)

	return &Blocker{
		blockerBase: newBlockerBase(
			threshold, interval, blockInterval, opts.maxSize, opts.gcParams),
		skip:   opts.skip,
		ipMask: opts.ipMask,
	}
}

func (b *Blocker) toKey(ip net.IP) key {
	return ipToKey(b.ipMask.Mask(ip))
}

// Allow returns true if the incoming request is allowed.
func (b *Blocker) Allow(now int64, ip net.IP) bool {
	if b.skip.Contains(ip) {
		return true
	}

	b.mx.Lock()
	defer b.mx.Unlock()

	return b.blockerBase.Allow(now, b.toKey(ip))
}

// Inc increments the rate limit counter for the specified peer. If
// this causes the peer to be blocked, the function will return true.
func (b *Blocker) Inc(now int64, ip net.IP) bool {
	if b.skip.Contains(ip) {
		return false
	}

	b.mx.Lock()
	defer b.mx.Unlock()

	return b.blockerBase.Inc(now, b.toKey(ip))
}
