package ratelimit

import (
	"encoding/binary"
	"math/rand"
	"testing"
	"time"
)

var rnd *rand.Rand

func init() {
	//nolint:gosec
	rnd = rand.New(rand.NewSource(42))
}

func testkey() key {
	var b [16]byte
	binary.BigEndian.PutUint64(b[:8], rnd.Uint64())
	binary.BigEndian.PutUint64(b[8:16], rnd.Uint64())
	return key(b)
}

func testkeys(n int) []key {
	keys := make([]key, 0, n)
	for i := 0; i < n; i++ {
		keys = append(keys, testkey())
	}
	return keys
}

func TestCounter_Cleanup(t *testing.T) {
	key1 := testkey()
	key2 := testkey()

	c := newCounter(10, 0, DefaultGCParams())
	c.Inc(0, key1)
	c.Inc(1, key1)
	cleanup(c, 2, time.Now().Add(1*time.Second))

	if n := c.Get(3, key1); n != 2 {
		t.Errorf("key1@3: n=%d, expected 2", n)
	}

	c.Inc(10, key1)
	c.Inc(99, key2)
	cleanup(c, 100, time.Now().Add(1*time.Second))

	if n := c.Get(101, key1); n != 0 {
		t.Errorf("key1@101: n=%d, expected 0", n)
	}
	if _, ok := c.byKey[key1]; ok {
		t.Errorf("key1@101: n=0 but the entry was not cleaned up")
	}

	if n := c.Get(101, key2); n != 1 {
		t.Errorf("key2@101: n=%d, expected 1", n)
	}
}

func BenchmarkCounter_Sparse(b *testing.B) {
	b.StopTimer()

	c := newCounter(10, 0, DefaultGCParams())

	b.StartTimer()

	for i := 0; i < b.N; i++ {
		t := int64(100 * i / b.N)
		c.Inc(t, testkey())
	}
}

func BenchmarkCounter_Dense(b *testing.B) {
	b.StopTimer()

	c := newCounter(10, 0, DefaultGCParams())

	keys := testkeys(100)

	b.StartTimer()

	for i := 0; i < b.N; i++ {
		t := int64(100 * i / b.N)
		c.Inc(t, keys[i%len(keys)])
	}
}

type tsEntry struct {
	t   int64
	key key
}

func timeseries(key key, duration, rate float64) []tsEntry {
	n := int(rate * duration)
	out := make([]tsEntry, n)
	for i := 0; i < n; i++ {
		t := int64(float64(i) * duration / float64(n))
		out[i] = tsEntry{t: t, key: key}
	}
	return out
}

func TestRatelimit(t *testing.T) {
	rl := newRatelimiterBase(10, 3, 0, DefaultGCParams())

	key1 := testkey()
	for _, entry := range timeseries(key1, 3.0, 20) {
		rl.rate.Inc(entry.t, entry.key)
	}

	key2 := testkey()
	for _, entry := range timeseries(key2, 3.0, 2) {
		rl.rate.Inc(entry.t, entry.key)
	}

	if ok := rl.Allow(3, key1); ok {
		t.Errorf("key1 (20qps) was not rate limited")
	}

	if ok := rl.Allow(6, key1); !ok {
		t.Errorf("key1 (20qps) rate limit did not expire")
	}

	if ok := rl.Allow(3, key2); !ok {
		t.Errorf("key2 (2qps) was rate limited")
	}
}

func TestBlocker(t *testing.T) {
	rl := newBlockerBase(10, 3, 100, 0, DefaultGCParams())

	key1 := testkey()
	for _, entry := range timeseries(key1, 3.0, 20) {
		rl.Inc(entry.t, entry.key)
	}

	key2 := testkey()
	for _, entry := range timeseries(key2, 3.0, 2) {
		rl.Inc(entry.t, entry.key)
	}

	if ok := rl.Allow(3, key1); ok {
		t.Errorf("key1 (20qps) was not blocked")
	}

	if ok := rl.Allow(6, key1); ok {
		t.Errorf("key1 (20qps) block expired too soon")
	}

	if ok := rl.Allow(600, key1); !ok {
		t.Errorf("key1 (20qps) block did not expire")
	}

	if ok := rl.Allow(3, key2); !ok {
		t.Errorf("key2 (2qps) was blocked")
	}
}
