package ratelimit

import (
	"errors"
	"fmt"
	"strconv"
	"strings"
	"time"
)

// ParseRate parses a rate expression of the form
// "<threshold>/<interval>" into its threshold and interval numeric
// parts. Threshold should be a non-zero integer, while the interval
// should be specified in a format that time.ParseDuration
// understands.
//
// Valid examples would be "1/5s", "10/1h", etc.
func ParseRate(s string) (int, int, error) {
	n := strings.IndexRune(s, '/')
	if n < 0 {
		return 0, 0, errors.New("syntax error")
	}

	threshold, err := strconv.ParseInt(s[:n], 10, 64)
	if err != nil {
		return 0, 0, fmt.Errorf("error parsing threshold: %w", err)
	}
	if threshold <= 0 {
		return 0, 0, errors.New("invalid threshold")
	}

	interval, err := time.ParseDuration(s[n+1:])
	if err != nil {
		return 0, 0, fmt.Errorf("error parsing interval: %w", err)
	}
	if interval <= 0 {
		return 0, 0, errors.New("invalid interval")
	}

	return int(threshold), int(interval.Seconds()), nil
}
